// $Id: jpcap.cpp,v 1.4 2008/10/06 15:08:18 rickwall Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/

// A libpcap wrapper with java native hooks.
//
// @author Jonas Lehmann and Patrick Charles
// @author Esmond Pitt
// @version $Revision: 1.4 $
// @lastModifiedBy $Author: rickwall $
// @lastModifiedAt $Date: 2008/10/06 15:08:18 $
//
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <exception>
//#include <ws2tcpip.h> This is needed for the cygwin compile.  Commenting out until I fix it properly

#ifdef WIN32
#define WPCAP
#define HAVE_REMOTE
#undef HAVE_SETDIRECTION
#else
// TODO define HAVE_SETDIRECTION for platforms where it exists.
#define HAVE_SETDIRECTION
#endif

extern "C" {
#include <pcap.h>
}
#include <jni.h>
#include "jnu.hpp"
#include "pcap.hpp"
#include "process.hpp"


#ifndef WIN32
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <unistd.h>
#else /*WIN32*/

/* KRIS: (28/04/2002) It seems to me that __int64 is a CygWin GCC problem
   MSVC has __int64 as a build type */
#ifdef	__GNUC__
//  typedef long __int64;
//  /* FIXME: This is dirty.            */
//  /* __int64 required for jlong in jni_md.h */
//  /* __int64 should be 8 byte integer         */

#endif
#include <Win32-Extensions.h>
#endif /*WIN32*/

#ifdef HAVE_REMOTE
static char RPCAP_LOCAL[] = "rpcap://";
#endif

// native function headers generated by javah
#include "net_sourceforge_jpcap_capture_PacketCapture.h"
#include "net_sourceforge_jpcap_capture_JWinPcap.h"

#if !defined(__FreeBSD__) && !defined(__NetBSD__) && !defined(__OpenBSD__) && \
    !defined(__bsdi__) && !defined(__APPLE__) && !defined(WIN32) && \
    !defined(__CYGWIN__)
# define DO_SELECT
#endif

/*****************************************************************************
 * global variables
 */

#define FALSE 0
#define TRUE 1

// Defined to as to agree with same variable in JWinPcap.java
#define DEFAULT_BUFFER_SIZE	net_sourceforge_jpcap_capture_JWinPcap_DEFAULT_BUFFER_SIZE

const int VERBOSE = net_sourceforge_jpcap_capture_PacketCapture_VERBOSE; // for debugging

#define TRACE(x) do { if (VERBOSE) { x; } } while (0)

const char DEVICE_NOT_FOUND_EXCEPTION[] =
  "net/sourceforge/jpcap/capture/CaptureDeviceNotFoundException";
const char DEVICE_OPEN_EXCEPTION[] =
  "net/sourceforge/jpcap/capture/CaptureDeviceOpenException";
const char DEVICE_LOOKUP_EXCEPTION[] =
  "net/sourceforge/jpcap/capture/CaptureDeviceLookupException";
const char CAPTURE_PACKET_EXCEPTION[] =
  "net/sourceforge/jpcap/capture/CapturePacketException";
const char CONFIGURATION_EXCEPTION[] =
  "net/sourceforge/jpcap/capture/CaptureConfigurationException";
const char INVALID_FILTER_EXCEPTION[] =
  "net/sourceforge/jpcap/capture/InvalidFilterException";
const char CLASS_EXCEPTION[] =
  "java/lang/ClassNotFoundException";
const char FILE_OPEN_EXCEPTION[] =
  "net/sourceforge/jpcap/capture/CaptureFileOpenException";
const char IO_EXCEPTION[] =
  "java/io/IOException";
const char EOF_EXCEPTION[] =
  "java/io/EOFException";
const char SOCKET_TIMEOUT_EXCEPTION[] =
  "java/net/SocketTimeoutException";
const char	UNSUPPORTED_OPERATION_EXCEPTION[] =
	"java/lang/UnsupportedOperationException";
const char	RUNTIME_EXCEPTION[] =
	"java/lang/RuntimeException";
const char	ILLEGAL_ARGUMENT_EXCEPTION[] =
	"java/lang/IllegalArgumentException";

/*****************************************************************************
 * utility functions
 */

/*
 * Write captured packet data to stderr.
 */
void printPacket(u_char *user, u_char *cp, u_int caplen, u_int length,
                 u_int seconds, u_int useconds)
{
  register u_int i, s;
  register int nshorts;

  nshorts = (u_int) caplen / sizeof(u_short);

  fprintf(stderr, "Packet: u = %s, l = %d of %d, t = %u.%06u, d = ",
          user, caplen, length, seconds, useconds);

  i = 0;
  while (--nshorts >= 0) {
    if ((i++ % 8) == 0)
      fprintf(stderr, "\n\t\t\t");
    s = *cp++;
    fprintf(stderr, " %02x%02x", s, *cp++);
  }

  if (caplen & 1) {
    if ((i % 8) == 0)
      (void)fprintf(stderr, "\n\t\t\t");
    (void)fprintf(stderr, " %02x", *cp);
  }

  fprintf(stderr, "\t\n");
}


/*
 * Create and throw an exception into the Java VM using this wrapper.
 * If the specified exception class can't be found, the VM will throw
 * a class not found exception.
 */
void throwException(JNIEnv *env, const char *excClassName, const char *message) {
	// Declare a local frame
//	JNULocalFrame	frame;
	// printf("throwing exception %s(%s)\n", excClassName, message); fflush(stdout);
	// create an instance of the specified exception
  jclass exception = env->FindClass(excClassName);

  if(exception != NULL)
    // throw the new exception back to the java wrapper
    env->ThrowNew(exception, message);

  // free the local reference. if exception is still null, delete is a noop.
  env->DeleteLocalRef(exception);

}

/*****************************************************************************
 * implementation of pcap wrapper functions with java hooks
 */

#if UNUSED

/*
 * Callback called by pcap_loop()/pcap_dispatch() for processing packet data.
 * This method thunks to the C++ PacketProcessor::processData()
 * method, which decomposes the captured data and transfers it
 * to the PacketCapture object in the attached java VM.
 */
void processData(u_char *user, const struct pcap_pkthdr *h, const unsigned char *sp) {
	// Thunk to C++ callback
	PacketProcessor *proc = (PacketProcessor *) user;
	proc->handlePacket(h, sp);
}

	// Now implemented in Java, ultimately calling findCaptureDevice0()
/*
 * Look up a network device.
 * Throws a CaptureDeviceNotFoundException if no suitable device is found.
 *
 * JNI PacketCapture.findDevice0()
 */
JNIEXPORT jstring JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_findDevice0
(JNIEnv *env, jobject self) {

	char ebuf[PCAP_ERRBUF_SIZE];

	char *device = pcap_lookupdev(ebuf);

	if(device == NULL) {
		// if no device can be found, throw an exception back to the java wrapper
		throwException(env, DEVICE_NOT_FOUND_EXCEPTION, ebuf);
		return NULL;
	}
#ifndef WIN32
	return env->NewStringUTF(device);
#else /*WIN32*/
	//char buf[256];
	char buf_all[516];
	char* desc = 0;
	int i=0, t=0;
	wchar_t *dev = (wchar_t *)device;
    if (dev[0]<256) { /* NT / 2000 */
		if (dev[0]==0 && dev[1]==0) {
			throwException(env, DEVICE_NOT_FOUND_EXCEPTION, ebuf);
			return NULL; /* no device */
		}
		// Do the same thing as lookupDevices, so that we can
		// find a human-readable device name
		//wcstombs(buf,(wchar_t *)dev,255);
		//return env->NewStringUTF(buf);

		while(!(dev[i]==0 && dev[i-1]==0)) {
			i++;
		}

		// human-readable descriptions are placed after the device names
		desc = (char*)(dev + i + 1);

		//      t = wcstombs(buf_all,(wchar_t *)dev,255);
        t = 0;
        while (dev[i] != 0) {
			if (t < sizeof buf_all/2)
				buf_all[t++] = dev[i];
			i++;	// unconditional increment! EJP 9 May 2008
        }
        i++;

#if BROKEN_WINDOWS_DESCRIPTION_MISFEATURE
		buf_all[t++] = '\n';
		buf_all[t++] = ' ';
		buf_all[t++] = ' '; // set "\n  " as windows signature

		while (*desc != 0) {
			if (t < sizeof(buf_all)-2)
				buf_all[t++] = *desc;
			desc++;
		}
#endif
		buf_all[t++] = '\0';	// Added EJP 30 April 2008
		return env->NewStringUTF(buf_all);

    } else { /* 9x */
		char *dev9x = (char *)dev;

		if (dev9x[0]==0 && dev9x[1]==0) {
			throwException(env, DEVICE_NOT_FOUND_EXCEPTION, ebuf);
			return NULL; /* no device */
		}

		return env->NewStringUTF((char *)(dev9x));
    }

#endif
}
#endif
/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    create0
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_create0
  (JNIEnv *env, jobject self)
{
	return (jlong)new PacketCapture();
}

/*
 * Creates a pcap_t structure without starting a capture.
  * If the open is successful, sets the global device pointer.
 *
 * JNI PacketCapture.createDead0()
 */
JNIEXPORT void JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_createDead0
  (JNIEnv *env, jobject self, jlong handle, jint linkType, jint snaplen)
{
	// Declare a local frame
	JNULocalFrame	frame;

	PacketCapture	*pc = (PacketCapture*)handle;
  	pc->setPD(pcap_open_dead(linkType, snaplen));

  	pc->stop();

	return;
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    destroy0
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_destroy0
  (JNIEnv *env, jobject self, jlong handle)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	delete pc;
//	printf("Deleted a PacketCapture instance\n"); fflush(stdout);
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    configureBlocking0
 * Signature: (IZ)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_configureBlocking0
  (JNIEnv *env, jobject self, jlong handle, jboolean isBlocking)
{
	char ebuf[PCAP_ERRBUF_SIZE];

	PacketCapture	*pc = (PacketCapture*)handle;
	if (pcap_setnonblock(pc->getPD(), !isBlocking, ebuf) < 0)
		throwException(env, DEVICE_NOT_FOUND_EXCEPTION, ebuf);
}

/*
 * Open a network device for data capture.
 * Throws an OpenException if the device cannot be opened.
 * If the open is successful, sets the global device pointer.
 *
 * JNI PacketCapture.open0()
 */
JNIEXPORT jint JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_open0
  (JNIEnv *env, jobject self, jlong handle, jstring jdevice, jint snaplen,
   jboolean promiscuous, jint timeout)
{
	// Declare a local frame
	JNULocalFrame	frame;

  if(jdevice == NULL) {
    // device can't be null; throw an exception back to java wrapper
    throwException(env, DEVICE_OPEN_EXCEPTION, "null device specified");
    return 0;
  }

//  const char *device = env->GetStringUTFChars(jdevice, 0);
  JNUString	device(jdevice);

  char ebuf[PCAP_ERRBUF_SIZE];

	PacketCapture	*pc = (PacketCapture*)handle;
  pc->setPD(pcap_open_live(device, snaplen, promiscuous, timeout, ebuf));
  if(pc->getPD() == NULL) {
    // if the open fails, throw an exception back to the java wrapper
    throwException(env, DEVICE_OPEN_EXCEPTION, ebuf);
    return 0;
  }

  pc->stop();

  // return the link type
  return pcap_datalink(pc->getPD());
}

/*
 * Open a tcpdump-formatted "savefile" for reading captured packets
 * as an alternative to capturing packets off the wire.
 *
 * JNI PacketCapture.openOffline0()
 */
JNIEXPORT jint JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_openOffline0
  (JNIEnv *env, jobject self, jlong handle, jstring jfileName)
{
	// Declare a local frame
	JNULocalFrame	frame;
  char ebuf[PCAP_ERRBUF_SIZE];

  JNUString	fileName(jfileName);

  PacketCapture	*pc = (PacketCapture*)handle;
  pc->setPD(pcap_open_offline(fileName, ebuf));

  if(pc->getPD() == NULL) {
    throwException(env, FILE_OPEN_EXCEPTION, ebuf);
    return 0;
  }

  pc->stop();

  // return the link type
  return pcap_datalink(pc->getPD());

}

/*
 * Capture packets.
 * Throws a CapturePacketException if a problem occurs during the capture.
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    capture0
 * Signature: (JIII)I
 */
JNIEXPORT jint JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_capture0
(JNIEnv *env, jobject self, jlong handle, jint count, jint mode, jint timeout)
{
	try
	{
		// Declare a local frame
		JNULocalFrame	frame;

		// make sure a device is open before allowing the capture session to start
		PacketCapture	*pc = (PacketCapture*)handle;
		if(pc->getPD() == NULL) {
			throwException(env, CAPTURE_PACKET_EXCEPTION,
				"a device must be open before capturing packets");
			return -1;
		}

		// EJP 7 May 2008: Find our callback method.
		// Do this once, not on every packet callback.
		// Should be somewhat faster ;-)
		JNUVoidMethod	method(
			JNUObject<jobject>(self)
			.getClass()
			.getMethod("handlePacket", "(IIIILjava/nio/ByteBuffer;)V")
			);
		// EJP 7 May 2008: fixed 1609568: memory leak,
		// by allocating this object on the stack.
		PacketProcessor p(env, self, method);

		// EJP 12 May 2008: removed the old pcap_loop() conditional code block.

		int	cnt = 0;
		pc->go();

#ifdef DO_SELECT
		if (pcap_file(pc->getPD()) == NULL) { // live capture
			struct timeval	st,
							*stp = &st;
			// EJP 7 May 2008: fixed 1226668.
			unsigned	fileno = pcap_fileno(pc->getPD());

			// EJP 7 May 2008: simplified and fixed this test.
			// Previously a count of zero also meant infinity, which is not
			// what the Javadoc says, and disagrees with the libcpap
			// semantics of pcap_loop()/pcap_dispatch(), which we are supposed
			// to be emulating.
			while (pc->isGoing() && (count < 0 || cnt < count))
			{
				int	captureCount = count < 0 ? -1 : count-cnt;
				fd_set fd_wait;
				FD_ZERO(&fd_wait);
				FD_SET(fileno, &fd_wait);
				// EJP 26 May 2008: use the timeout specified at 'open' time
				// for selecting as well as looping, rather than 1s.
				if (timeout > 0)
				{
					st.tv_sec	 = timeout / 1000;
					st.tv_usec	= (timeout % 1000) * 1000;
				}
				else
					stp = NULL;

				int sel = select(FD_SETSIZE, &fd_wait, NULL, NULL, stp);

				if (sel == 0 && mode == net_sourceforge_jpcap_capture_PacketCapture_DISPATCH_MODE)	// timeout
					return cnt;

				if (sel > 0) { // OK, select says we have a packet in the queue
					// NB the cast to PcapCallback* is necessary so we don't lie to the thunk
					// about the type & address of the class object.
					int rc = pc->capture(captureCount, PcapCallback::pcap_handler_thunk, (u_char *)(PcapCallback*)&p);
					switch (rc)
					{
					case -2:	// loop stopped before any packets captured
						return cnt > 0 ? cnt : -2;
					case -1:	// error
						// if the capture failed, throw an exception back to the java wrapper
						throwException(env, CAPTURE_PACKET_EXCEPTION, pcap_geterr(pc->getPD()));
						return rc;
					case 0:	// no packets captured
						return cnt;
					default: // >= 0: # packets captured
						cnt += rc;
						break;
					}
				} else if (sel < 0) {
					throwException(env, CAPTURE_PACKET_EXCEPTION, "select() returns -1");
					return -1;
				}
			}
			return cnt;
		}
		else {
			// don't try to select on a file ...

#endif // DO_SELECT
			// EJP 12 May 2008: simplified this considerably.
			// 1. Removed the code duplication by altering the conditional compilation
			// logic, so the same code is now used for captures from files and devices.
			// 2. Pass the application's count rather than iterating 1 at a time.
			// 3. Call either pcap_loop() or pcap_dispatch(),
			//		depending on the loop mode, but with a correct count,
			//		so as to letting pcap_loop() implement the former outer logic loop.
			// 4. Handled rc == -2 correctly.

			// NB the cast to PcapCallback* is necessary so we don't lie to the thunk
			// about the type & address of the class object.
			int rc = pc->capture(count, PcapCallback::pcap_handler_thunk, (u_char *)(PcapCallback*)&p);
			switch (rc)
			{
			case -1:	// error
				// if the capture failed, throw an exception back to the java wrapper
				throwException(env, CAPTURE_PACKET_EXCEPTION, pcap_geterr(pc->getPD()));
				return rc;
			case 0:	// no packets captured
				if (pcap_file(pc->getPD()) != NULL)
				{
					// We are capturing from a file so we throw an EOFException
					throwException(env, EOF_EXCEPTION, pcap_geterr(pc->getPD()));
				}
				// fall through
			case -2:	// loop stopped before any packets captured
			default: // > 0: # packets captured
				return rc;
			}
#ifdef DO_SELECT
		}
#endif // DO_SELECT
	}
	catch (PendingJavaException &)
	{
		return 0;
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    captureNext0
 * Signature: (JLjava/nio/ByteBuffer;)I
 */
JNIEXPORT jlong JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_captureNext0
  (JNIEnv *env, jobject self, jlong handle, jobject jbb)
{
	try
	{
		// Declare a local frame
		JNULocalFrame	frame;
		char ebuf[PCAP_ERRBUF_SIZE];
		PacketCapture	*pc = (PacketCapture*)handle;
		struct pcap_pkthdr *pkt_header;
		const u_char	*pkt_data;

		int	rc = pcap_next_ex(pc->getPD(), &pkt_header, &pkt_data);
//		fprintf(stdout, "captureNext0(): pcap_next_ex() rc=%d\n", rc); fflush(stdout);
		switch (rc)
		{
		case -1:
			throwException(env, CAPTURE_PACKET_EXCEPTION, pcap_geterr(pc->getPD()));
			return 0;
		case -2:	// EOF on savefile
			throwException(env, EOF_EXCEPTION, pcap_geterr(pc->getPD()));
			return 0;
		case 0:
			// timeout
			throwException(env, SOCKET_TIMEOUT_EXCEPTION, pcap_geterr(pc->getPD()));
			return 0;
		case 1:
			{
				// Declare a local frame
//				JNULocalFrame	frame;
				JNUByteBuffer	bb(jbb);
				// Sanity checks. If these fail, the Java wrapper code is wrong.
				if (bb.getLimit() < pkt_header->caplen)
				{
					throwException(env, RUNTIME_EXCEPTION, "maximum capture length exceeded");
					return 0;
				}
				memcpy(bb.getData(), pkt_data, pkt_header->caplen);
				bb.setLimit(pkt_header->caplen);
				bb.commit();
			}
			// Return the timestamp
			return pkt_header->ts.tv_sec*1000000L+pkt_header->ts.tv_usec;
		default:
			sprintf(ebuf, "Unexpected return value from pcap_next_ex(): %d\n", rc);
			throwException(env, RUNTIME_EXCEPTION, ebuf);
			return 0;
		}
	}
	catch (PendingJavaException &)
	{
		return 0;
	}
}

/*
 * Stop capturing packets
 *
 * JNI PacketCapture.endCapture0()
 */
JNIEXPORT void JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_endCapture0
(JNIEnv *env, jobject self, jlong handle) {
  PacketCapture	*pc = (PacketCapture*)handle;
  pc->stop();
  // EJP 12 May 2008: call this to stop pcap_loop()/pcap_dispatch() internally.
  pcap_breakloop(pc->getPD());
}

#if UNUSED	// No longer calls native code
/*
 * Fetch the network associated with the device.
 * Throws a CaptureConfigurationException if the device is messed up.
 *
 * JNI PacketCapture.getNetwork0()
 */
JNIEXPORT jint JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_getNetwork0
(JNIEnv *env, jobject self, jstring jdevice) {
  char ebuf[PCAP_ERRBUF_SIZE];
  bpf_u_int32 localnet = 0;
  bpf_u_int32 netmask = 0;

//  const char *device = env->GetStringUTFChars(jdevice, 0);
  StringUTFChars	device(env, jdevice);

  int rc = pcap_lookupnet(const_cast<char*>(device), &localnet, &netmask, ebuf);
  if (rc < 0) {
    // if the lookup failed, throw an exception back to the java wrapper
    throwException(env, CONFIGURATION_EXCEPTION, ebuf);
    return 0;
  }
  else
    return localnet;
}

/*
 * Fetch the netmask associated with the device.
 * Throws a CaptureConfigurationException if the device is messed up.
 *
 * JNI PacketCapture.getNetmask0()
 */
JNIEXPORT jint JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_getNetmask0
(JNIEnv *env, jobject self, jstring jdevice) {
  char ebuf[PCAP_ERRBUF_SIZE];
//  const char *device = env->GetStringUTFChars(jdevice, 0);
  StringUTFChars	device(env, jdevice);
  bpf_u_int32 localnet = 0;
  bpf_u_int32 netmask = 0;

  int rc = pcap_lookupnet(const_cast<char*>(device), &localnet, &netmask, ebuf);
  if (rc < 0) {
    // if the lookup failed, throw an exception back to the java wrapper
    throwException(env, CONFIGURATION_EXCEPTION, ebuf);
    return 0;
  }
  else
    return netmask;
}
#endif	// UNUSED

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    getLinkLayerDescription0
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_getLinkLayerDescription0
  (JNIEnv *env, jobject self, jint type)
{
	try
	{
		return JNUString(pcap_datalink_val_to_description(type));
	}
	catch (PendingJavaException &)
	{
		return 0;
	}

}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    getLinkLayerName0
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_getLinkLayerName0
  (JNIEnv *env, jobject self, jint linkType)
{
	try
	{
		return JNUString(pcap_datalink_val_to_name(linkType));
	}
	catch (PendingJavaException &)
	{
		return 0;
	}
}

/*
 * Fetch the link layer type associated with this device.
 *
 * JNI PacketCapture.getLinkLayerType()
 */
JNIEXPORT jint JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_getLinkLayerType0
(JNIEnv *env, jobject self, jlong handle) {
  PacketCapture	*pc = (PacketCapture*)handle;
  return pcap_datalink(pc->getPD());
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    getLinkLayerTypes0
 * Signature: (I)[I
 */
JNIEXPORT jintArray JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_getLinkLayerTypes0
  (JNIEnv *env, jobject self, jlong handle)
{
	try
	{
		// Declare a local frame
		JNULocalFrame	frame;
		PacketCapture	*pc = (PacketCapture*)handle;
		int	*types;
		int rc = pcap_list_datalinks(pc->getPD(), &types);
		if (rc < 0)
		{
			throwException(env, CONFIGURATION_EXCEPTION, pcap_geterr(pc->getPD()));
			return 0;
		}
		jint	*jintTypes = new jint[rc];
		for (int i = 0; i < rc; i++)
			jintTypes[i] = types[i];
	//	jintArray	result = env->NewIntArray(rc);
	//	env->SetIntArrayRegion(result, 0, rc, jintTypes);
		JNUIntArray	result(rc);
		result.setRegion(jintTypes, 0, rc);
		delete jintTypes;
		/*
		 * The libcap API documentation says:
		 * 'pcap_list_datalinks() allocates an array to hold the list and sets *dlt_buf.
		 * The caller is responsible for freeing the array.'
		 * But of course it doesn't say *how* to free the array,
		 * and of course uncommenting the following line causes
		 * the usual free()/wrong DLL/C library problem.
		 * So this is a memory leak.
		 */
		// free(types);
		return frame.pop((jintArray)result);
	}
	catch (PendingJavaException &)
	{
		return 0;
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    setLinkLayerType0
 * Signature: (II)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_setLinkLayerType0
  (JNIEnv *env, jobject self, jlong handle, jint linkType)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	if (pcap_set_datalink(pc->getPD(), linkType) < 0)
	    throwException(env, CONFIGURATION_EXCEPTION, pcap_geterr(pc->getPD()));
}

/*
 * Create and compile a filter expression.
 *
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    compileFilter0
 * Signature: (IIILjava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_compileFilter0
  (JNIEnv *env, jobject self, jlong handle, jboolean optimize, jint netmask, jstring jfilter)
{
	PacketCapture	*pc = (PacketCapture*)handle;
//	char *filter = const_cast<char*>(env->GetStringUTFChars(jfilter, 0));
	JNUString	filter(jfilter);

  // EJP 13 May 2008: pass the correct netmask so that broadcast filters can work
	try
	{
		return (jlong)new PcapCompiledFilter(pc->getPD(), filter, optimize, htonl(netmask));
	}
	catch (const PcapFilterException &exc)
	{
		throwException(env, INVALID_FILTER_EXCEPTION, exc.getMessage());
		return 0;
	}
}

/*
 * Free a compiled filter program.
 *
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    freeFilter0
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_freeFilter0
  (JNIEnv *env, jobject self, jlong handle)
{
	// EJP 12 May 2008: yet another memory leak fixed.
	// Free the BPF data
	if (handle != 0)
	{
		PcapCompiledFilter	*pcf = (PcapCompiledFilter*)handle;
		delete pcf;
	}
}

/*
 * Activate a filter from a compiled filter program.
 *
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    setFilter0
 * Signature: (IJ)V
 */
JNIEXPORT void JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_setFilter0
(JNIEnv *env, jobject self, jlong handle, jlong filter) {
	PacketCapture	*pc = (PacketCapture*)handle;

	// printf("setFilter0() handle=%lx filter=%lx\n", handle, filter);

	// the device must be open in order to set the filter.
	if(pc->getPD() == NULL) {
		throwException(env, INVALID_FILTER_EXCEPTION,
			"A device must be open before setting the filter.");
		return;
	}

	// activate program
	bpf_program *bpp = ((PcapCompiledFilter*)filter)->getFilter();
	if(pcap_setfilter(pc->getPD(), bpp) == -1)
		// if the filter couldn't be activated, throw an exception back to java
		throwException(env, INVALID_FILTER_EXCEPTION, pcap_geterr(pc->getPD()));

}

#if UNUSED	// EJP: now remembering this at the Java level, much simpler.
/*
 * Return the snapshot length being used (given network device has been
 * opened) by using a pcap library API call.
 *
 * JNI PacketCapture.getSnapshotLength()
 */
JNIEXPORT jint JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_getSnapshotLength0
  (JNIEnv *env, jobject self, jlong handle)
{
  if(pc->getPD() != NULL) {
    return pcap_snapshot(pc->getPD());
  }
  return -1;
}
#endif

/*
 * Close the capture device.
 *
 * JNI PacketCapture.close0()
 */
JNIEXPORT void JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_close0
(JNIEnv *env, jobject self, jlong handle) {
	PacketCapture	*pc = (PacketCapture*)handle;
	pc->closePD();
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    closeDump0
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_closeDump0
  (JNIEnv *env, jobject self, jlong handle)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	pc->closeDumper();
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    flushDump0
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_flushDump0
  (JNIEnv *env, jobject self, jlong handle)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	if (pcap_dump_flush(pc->getDumper()) < 0)
	{
		throwException(env, IO_EXCEPTION, pcap_geterr(pc->getPD()));
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    ftellDump0
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_ftellDump0
  (JNIEnv *env, jobject self, jlong handle)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	long	rc = pcap_dump_ftell(pc->getDumper());
	if (rc < 0)
	{
		throwException(env, IO_EXCEPTION, pcap_geterr(pc->getPD()));
	}
	return rc;
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    openDump0
 * Signature: (ILjava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_openDump0
  (JNIEnv *env, jobject self, jlong handle, jstring jfilename)
{
	try
		{
			PacketCapture	*pc = (PacketCapture*)handle;
			if (pc->getPD()==NULL)
			{
		    	throwException(env, DEVICE_OPEN_EXCEPTION, "pcap_t not initialized, call createDump0 first.");
	    		return;
			}

			JNUString	filename(jfilename);
			pc->setDumper(pcap_dump_open(pc->getPD(), filename));
			if (pc->getDumper() == 0)
			{
				throwException(env, IO_EXCEPTION, pcap_geterr(pc->getPD()));
			}
		}
	catch (PendingJavaException &)
	{
		return;
	}
}

/*
 * Fetch and set statistics in the java vm.
 *
 * JNI PacketCapture.getStatistics()
 */
JNIEXPORT jobject JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_getStatistics0
(JNIEnv *env, jobject self, jlong handle)
{
	try
	{
	  struct pcap_stat stat =
	  // Need to initialize these to zero, MSVC defies the C++ rules on this.
	  {
		  0,	// ps_recv
		  0,	// ps_drop
		  0,	// ps_ifdrop
#ifdef HAVE_REMOTE
		  0,	// ps_capt
		  0,	// ps_sent
		  0,	// ps_netdrop
#endif
	  };

	  PacketCapture	*pc = (PacketCapture*)handle;
	  if (pc->getPD() != NULL && pcap_file(pc->getPD()) == NULL) {
#ifndef WIN32
		  pcap_stat	*pcs = &stat;
		  if (pcap_stats(pc->getPD(), &stat) < 0)
#else
		  int	len;
		  pcap_stat	*pcs = pcap_stats_ex(pc->getPD(), &len);
		  if (pcs == 0)
#endif
		  {
		  /*
		   * todo: throw the following as an exception..
		   fprintf(stderr, "jpcap: pcap_stats(%s) %s\n",
		 g_device, pcap_geterr(pc->getPD()));
		  */
	//		fprintf(stderr, "jpcap: pcap_stats() %s\n", pcap_geterr(pc->getPD()));
			throwException(env, RUNTIME_EXCEPTION, pcap_geterr(pc->getPD()));
		}
		else {
			JNUClass	cls(JNUClass::forName("net/sourceforge/jpcap/capture/CaptureStatistics"));
			JNUConstructor	cid(cls.getConstructor("(IIIIII)V"));

		  // received count
			int	received			= pcs->ps_recv;
		  // dropped count
			int dropped				= pcs->ps_drop;
		/* drops by interface XXX not yet supported */
			int droppedByInterface	= pcs->ps_ifdrop;
#ifdef HAVE_REMOTE
			/* number of packets that are received by the application */
			int receivedByApplication	= pcs->ps_capt;
			/* number of packets sent by the server on the network */
			int sentByServer			= pcs->ps_sent;
			/* number of packets lost on the network */
			int droppedByNetwork		= pcs->ps_netdrop;
#else
			/* number of packets that are received by the application */
			int receivedByApplication	= 0;
			/* number of packets sent by the server on the network */
			int sentByServer		= 0;
			/* number of packets lost on the network */
			int droppedByNetwork	= 0;
#endif
			// printf("received=%d, dropped=%d, droppedByInterface=%d\n", received, dropped, droppedByInterface);
			jobject capture_stats = cid.newInstance(cls, received, dropped, droppedByInterface, receivedByApplication, sentByServer, droppedByNetwork);
		  // free local storage for 'cls'.
		  // env->DeleteLocalRef(cls);
			return capture_stats;
		}
	//	  printf("receivedCount=%d droppedCount=%d\n", stat.ps_recv, stat.ps_drop);
	  }
	  return 0;
	}
	catch (PendingJavaException &)
	{
		return 0;
	}
}

#if UNUSED
/*
 * This function is currently unused. Might be used later by capture
 * to potentially adjust the snapshot length in the event that the
 * capture client got it wrong.
 *
 * EJP: I don't see how. You can't get at the snaplen that libpcap is using,
 * except by closing & reopening the device, so what on earth is the point?
 * Deleted this.
 */
#endif

#if UNUSED
// Now implemented in Java, ultimately calling lookupCaptureDevices()
/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    lookupDevices0
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_lookupDevices0
(JNIEnv *env, jclass cls)
{
	jobjectArray	devices = NULL;
  char ebuf[PCAP_ERRBUF_SIZE];

#ifndef WIN32
  int sock=socket(AF_INET,SOCK_DGRAM,0);
  struct ifconf ifc;
  struct ifreq *ifr,*last;
  struct ifreq ifrflags;
  pcap_t *pch;

  char names[100][100];
  int total = 0, i = 0, ifrSize = 0;

  if(sock < 0 ){
    /* error opening socket */
    throwException(env, DEVICE_LOOKUP_EXCEPTION, strerror(sock));
    return NULL;
  }

  ifc.ifc_len = 1024 * sizeof(struct ifreq);
  ifc.ifc_buf = (char *)malloc(ifc.ifc_len);

  if(ioctl(sock,SIOCGIFCONF,&ifc) < 0 ||
     ifc.ifc_len < (int)sizeof(struct ifreq)){
    /* SIOCGIFCONF error */
    goto FAIL;
  }

  ifr = (struct ifreq *)ifc.ifc_req;
  last = (struct ifreq *)((char *)ifr+ifc.ifc_len);

#ifdef HAVE_SA_LEN
    ifrSize = ifr->ifr_addr.sa_len+IFNAMSIZ;
#else /* HAVE_SA_LEN */
    ifrSize = sizeof(struct ifreq);
#endif /* HAVE_SA_LEN */

  for(;ifr < last; (char*)ifr += ifrSize) {
    /* Skip "dummy" and "alias" interface */
    /*
    if(strncmp(ifr->ifr_name,"dummy",5)==0 ||
       strchr(ifr->ifr_name,':') != NULL)
      continue;
    */

    for(i=0;i<total;i++)
      if(strcmp(names[i],ifr->ifr_name)==0)
	continue;

    /* Check flags */
    memset(&ifrflags,0,sizeof ifrflags);
    strncpy(ifrflags.ifr_name,ifr->ifr_name,sizeof ifrflags.ifr_name);
    if(ioctl(sock,SIOCGIFFLAGS,(char *)&ifrflags)<0){
      if(errno == ENXIO)
        continue;
      else
        goto FAIL;
    }

    if(!(ifrflags.ifr_flags & IFF_UP))
      continue;

    pch = pcap_open_live(ifr->ifr_name,68,0,0, ebuf);
    if(pch == NULL)
      continue;
    pcap_close(pch);
    strcpy(names[total++],ifr->ifr_name);
  }

  if( total > 0 )  {
	  devices = env->NewObjectArray
		  ((jsize)total, JNUClass(env, "java/lang/String"), NULL);
	  env->ExceptionDescribe();
	  if (env->ExceptionCheck())
		  return 0;
	  for(i = 0 ; i < total ; i++ ) {
		  env->SetObjectArrayElement(devices, i, env->NewStringUTF(names[i]));
		  env->ExceptionDescribe();
		  if (env->ExceptionCheck())
			  return 0;
	  }
  }

FAIL:
  free(ifc.ifc_buf);
  close(sock);

#else /* WIN32 */
  wchar_t *dev;
  int i=0,c=0,j=0, t=0; // t = array index
  //char buf[256];
  //char buf2[256];
  char buf_all[516];
  char* desc = 0;
  dev = (wchar_t *)pcap_lookupdev(ebuf);
  if(dev){
    if(dev[0]<256) { /* NT / 2000 */
      if(dev[0]==0 && dev[1]==0)
        return NULL; /* no device */
      while(!(dev[i]==0 && dev[i-1]==0)){
        if(dev[i]==0) c++;
        i++;
      }

      // human-readable descriptions are placed after the device names
      desc = (char*)(dev + i + 1);

      devices=env->NewObjectArray
	((jsize)c, JNUClass(env, "java/lang/String"), NULL);
      //descriptions=env->NewObjectArray
      //  ((jsize)c, JNUClass(env, "java/lang/String"), NULL);
	  env->ExceptionDescribe();
	  if (env->ExceptionCheck())
		  return 0;

      i = 0;

      for(j=0; j<c; j++){
      /*
	//An attempt to set description strings within an array that was
	//passed as an argument .. still not sure how to do this.

        wcstombs(buf,(wchar_t *)(dev + i), 255);
        env->SetObjectArrayElement(devices, j, env->NewStringUTF(buf));
        wcstombs(buf2,(wchar_t *)(desc + t), 255);
        env->SetObjectArrayElement(descriptions, j, env->NewStringUTF(buf2));

        while(dev[i]!=0) {
          i++;
        }
        i++;

        while(desc[t] != 0) {
          t++;
        }
        t++;
       */

        // This code will get the device description as well as dev name
        // and put them both in the devices array
        t = 0;
        while (dev[i] != 0) {
          if (t < sizeof buf_all/2)
            buf_all[t++] = dev[i];
		  i++;	// unconditional increment! EJP 9 May 2008
        }
        i++;

#if BROKEN_WINDOWS_DESCRIPTION_MISFEATURE
        buf_all[t++] = '\n';
        buf_all[t++] = ' ';
        buf_all[t++] = ' '; // set "\n  " as windows signature

        while (*desc != 0) {
          if (t < sizeof(buf_all)-2)
	    buf_all[t++] = *desc;
          desc++;
        }
#endif
        buf_all[t++] = '\0';
        env->SetObjectArrayElement(devices, j, env->NewStringUTF(buf_all));
		env->ExceptionDescribe();
		if (env->ExceptionCheck())
			return 0;
        desc++;
      }
    }else{ /* 9x */
      char *dev9x=(char *)dev;

      if(dev9x[0]==0 && dev9x[1]==0)
        return NULL; /* no device */
      while(!(dev9x[i]==0 && dev9x[i-1]==0)){
        if(dev9x[i]==0) c++;
        i++;
      }

      devices=env->NewObjectArray
		  ((jsize)c, JNUClass(env, "java/lang/String"), NULL);
	  env->ExceptionDescribe();
	  if (env->ExceptionCheck())
		  return 0;
      i=0;
      for(j=0;j<c;j++){
		  env->SetObjectArrayElement
			  (devices, j, env->NewStringUTF((char *)(dev9x + i)));
		  env->ExceptionDescribe();
		  if (env->ExceptionCheck())
			  return 0;
		  while(dev9x[i]!=0) i++;
		  i++;
      }
    }
  }
#endif /* WIN32 */

  if(devices == NULL) {
      devices=env->NewObjectArray
		  (0, JNUClass(env, "java/lang/String"), NULL);
	  env->ExceptionDescribe();
	  if (env->ExceptionCheck())
		  return 0;
  }
  return devices;
}
#endif

/*
* Class: net_sourceforge_jpcap_capture_PacketCapture
* Method: send0
* Signature: ()[Ljava/lang/String;
*/
JNIEXPORT jint JNICALL
Java_net_sourceforge_jpcap_capture_PacketCapture_send0
(JNIEnv *env, jobject self, jlong handle, jobject bb)
{
	try
	{
		// Declare a local frame
		JNULocalFrame	frame;
		PacketCapture	*pc = (PacketCapture*)handle;
		JNUByteBuffer	jbb(bb);
		jlong	length = jbb.getLimit();
		// Get a pointer to the data to send.
		const void	*data = jbb.getData();
#ifndef WIN32
		jint	ret = pcap_inject(pc->getPD(), (u_char*)data, (int)length);
#else	// WIN32
		jint	ret = pcap_sendpacket(pc->getPD(), (u_char*)data, (int)length);
#endif // WIN32
		if (ret < 0)
			throwException(env, IO_EXCEPTION, pcap_geterr(pc->getPD()));
		return ret;
	}
	catch (PendingJavaException &)
	{
		return 0;
	}
}

/*
 * Turn an IPv4 or IPv6 address into a Java byte array.
 * If the address pointer is null, a byte array containing the correct number
 * of zeroes for the address family is still returned.
 */
static jbyteArray getAddressByteArray(JNIEnv *env, int sa_family, struct sockaddr *addr)
{
	jbyteArray array = 0;
	int	len;
	switch (sa_family)
	{
	case AF_INET:
		len = sizeof(struct in_addr);
		array = env->NewByteArray(len);
		if (addr != 0)
			env->SetByteArrayRegion(array,0,len,(jbyte *)&((struct sockaddr_in *)addr)->sin_addr);
		break;
	case AF_INET6:
#ifndef WIN32
		// TODO check this for other platforms
		len = sizeof(struct in6_addr);
#else
		len = sizeof(struct in_addr6);
#endif
		array = env->NewByteArray(len);
		if (addr != 0)
			env->SetByteArrayRegion(array,0,len,(jbyte *)&((struct sockaddr_in6 *)addr)->sin6_addr);
		break;
	default:
		//printf("sa_family=%d\n",addr->sa_family);
		break;
	}
	env->ExceptionDescribe();
	if (env->ExceptionCheck())
		return 0;
	return array;
}

// Worker for both lookupCaptureDevices() methods
static jobjectArray	buildCaptureDevices(JNIEnv *env, pcap_if_t *alldevs, bool isExtended)
{
	// Declare a local frame
	JNULocalFrame	frame;
	char ebuf[PCAP_ERRBUF_SIZE];
	int	rc;
	// Count the devices
	jsize	count = 0;
	// Stupid MSVC requires this out here, otherwise code would be incompatible
	pcap_if_t *pcap_if;
	for (pcap_if = alldevs; pcap_if != 0; pcap_if = pcap_if->next)
	{
		// printf("device %s addresses=%x flags=%x\n", pcap_if->name, pcap_if->addresses, pcap_if->flags);
		// Test for device down, i.e. addresses = 0, ...
		if (pcap_if->addresses == 0)
			// => device is down, don't count it
			continue;

		// See if it can be opened.
		pcap_t	*pch;
#ifdef WIN32
		if (isExtended)
			pch = pcap_open(pcap_if->name, 0, 0, 0, 0, ebuf);
		else
#endif
			pch = pcap_open_live(pcap_if->name, 0, 0, 0, ebuf);
		if (pch == 0)
			// Can't open it, not much point in returning it to the user.
			continue;
		pcap_close(pch);

		// Device is OK, count it.
		count++;
	}
	// printf("found %d devices\n", count); fflush(stdout);

	// Find the required classes
	try
	{
		JNUClass	deviceClass(JNUClass::forName("net/sourceforge/jpcap/capture/CaptureDevice"));
		JNUClass	addressClass(JNUClass::forName("net/sourceforge/jpcap/capture/CaptureDevice$Address"));

		 // printf("got the classes\n"); fflush(stdout);
		// Get the CaptureDevice constructor
		JNUConstructor	deviceCtor(deviceClass.getConstructor(
			"("
			"Ljava/lang/String;"	// name
			"Ljava/lang/String;"	// description
			"III"					// network, netmask, flags
			"[Lnet/sourceforge/jpcap/capture/CaptureDevice$Address;"	// addresses
			")V"
			));
		// Get the Address constructor
		JNUConstructor addressCtor(addressClass.getConstructor("([B[B[B[B)V"));

		// printf("got the CaptureDevice ctor\n"); fflush(stdout);
		// Allocate the result array
		JNUObjectArray	deviceArray(count, deviceClass, 0);
		// printf("got the CaptureDevice[] array\n"); fflush(stdout);

		// Populate it
		int	item = 0;
		for (pcap_if = alldevs; pcap_if != 0; pcap_if = pcap_if->next)
		{
			// Push a new local frame so as to free the local references
			// to each CaptureDevice object ASAP.
			JNULocalFrame	frame;

			// Test for device down, i.e. addresses = 0, ...
			// printf("device %s addresses=%x\n", pcap_if->name, pcap_if->addresses); fflush(stdout);
			if (pcap_if->addresses == 0)
				// => device is down, don't return it
				continue;

			// See if it can be opened.
			pcap_t	*pch;
#ifdef WIN32
			if (isExtended)
				pch = pcap_open(pcap_if->name, 0, 0, 0, 0, ebuf);
			else
#endif
				pch = pcap_open_live(pcap_if->name, 0, 0, 0, ebuf);
			// printf("pch=%8.8x\n", pch); fflush(stdout);
			if (pch == 0)
				// Can't open it, not much point in returning it to the user.
				continue;
			pcap_close(pch);

			// Get the net and netmask properties
			// Very important to zero them first each time, that's
			// why they are redeclared.

			bpf_u_int32	network = 0,
						netmask = 0;
#ifdef HAVE_REMOTE
			int	type = -1;
			char host[PCAP_BUF_SIZE];
			char port[PCAP_BUF_SIZE];
			char name[PCAP_BUF_SIZE];
			// Get the local name that can be used by the old pcap functions
			rc = pcap_parsesrcstr(pcap_if->name, &type, host, port, name, ebuf);
			if (rc < 0) {
				throwException(env, CONFIGURATION_EXCEPTION, ebuf);
				return 0;
			}
#else
			char	*name = pcap_if->name;
#endif

			rc = pcap_lookupnet(name, &network, &netmask, ebuf);
			if (rc < 0) {
				// printf("pcap_lookupnet rc=%d ebuf=%s\n", rc, ebuf); fflush(stdout);
				// if the lookup failed, throw an exception back to the java wrapper
				throwException(env, CONFIGURATION_EXCEPTION, ebuf);
				return 0;
			}
			// NB network and netmask need to be adjusted to host byte order,
			// but in any case they are only 'int's so can't handle IPv6,
			// and they don't allow for NICs with multiple IP addresses,
			// so they are omitted from the public interface of CaptureDevice$Address.
			network = ntohl(network);
			netmask = ntohl(netmask);

			// printf("Device %s network=%8.8x flags=%x addresses=%8.8x\n", pcap_if->name, network, pcap_if->flags, pcap_if->addresses);

			jobject	description = 0;
			if (pcap_if->description)
			{
				// printf("\tdescription=%s\n", pcap_if->description); fflush(stdout);
				description = env->NewStringUTF(pcap_if->description);
			}

			// Get address-list properties

			// Count the address items
			int addrCount = 0;
			// Stupid MSVC requires this out here, otherwise code would be incompatible
			pcap_addr_t *addr;
			for (addr = pcap_if->addresses; addr != 0; addr = addr->next)
			{
				switch (addr->addr->sa_family)
				{
				case AF_INET:
				case AF_INET6:
					break;
				default:
					// printf("Unknown sa_family=%d\n", addr->addr->sa_family);
					// Skip unknown address types.
					continue;
				}
				addrCount++;
			}

			// Allocate the Address[] array
			JNUObjectArray	addressArray(addrCount, addressClass, NULL);

			// printf("\tAllocated the Address[] array\n"); fflush(stdout);

			int	addressItem = 0;
			for (addr = pcap_if->addresses; addr != 0; addr = addr->next)
			{
				// Push a new local frame so as to free the local references
				// to the addressObject and its constituent parts ASAP
				JNULocalFrame	frame;
				jbyteArray	address = 0,
							netmask = 0,
							broadcast = 0,
							destination = 0;
				int	sa_family = addr->addr->sa_family;
				switch (sa_family)
				{
				case AF_INET:	// IPv4
				case AF_INET6:	// IPv6
					address		= getAddressByteArray(env, sa_family, addr->addr);
					netmask		= getAddressByteArray(env, sa_family, addr->netmask);
					broadcast	= getAddressByteArray(env, sa_family, addr->broadaddr);
					destination	= getAddressByteArray(env, sa_family, addr->dstaddr);
					break;

				default:
	//				printf("Unknown sa_family=%d\n", sa_family);
					// Skip unknown address types.
					continue;
				}
				// Allocate the new address object
				jobject	addressObject = addressCtor.newInstance(addressClass, address, netmask, broadcast, destination);
				// printf("\tAllocated the Address object\n"); fflush(stdout);

				// Put it into the result array
				addressArray.set(addressItem++, addressObject);
	//			env->SetObjectArrayElement(addresses, addressItem++, addressObject);
				// printf("\tSetObjectArrayElement() has returned\n"); fflush(stdout);
			}

			// Allocate the new object
#ifdef HAVE_REMOTE
			// printf("device=%s type=%d host=%s port=%s name=%s\n", pcap_if->name, type, host, port, name);
	//		switch (type)
	//		{
	//		case PCAP_SRC_IFREMOTE:
				// If it really is remote, use the rpcap:// style name,
			if (isExtended)
				strcpy(name, pcap_if->name);
	//			break;
	//		default:
				// otherwise use the local name,
				// so as not to break absolutely everything.
	//			break;
	//		}
#endif
			JNUString	jstringName(name);
			jobject	cd = deviceCtor.newInstance
				(
				deviceClass,
				jstringName.object(),	// env->NewStringUTF(name),
				description,
				network,
				netmask,
				pcap_if->flags,
				(jobjectArray)addressArray
				);
			// printf("got the CaptureDevice instance count=%d item=%d\n", count, item); fflush(stdout);

			// Put it into the result array
			deviceArray.set(item++, cd);
	//		env->SetObjectArrayElement(devices, item++, cd);
			// printf("SetObjectArrayElement() has returned\n"); fflush(stdout);
		}

		if (strlen(ebuf) > 0)
		{
			TRACE(printf("buildCaptureDevices(): warning %s\n", ebuf));
		}

		return frame.pop((jobjectArray)deviceArray);
	}
	catch (const PendingJavaException &)
	{
		// printf("pending Java exception\n"); fflush(stdout);
		return 0;
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    lookupCaptureDevices0
 * Signature: (Ljava/lang/String;I)[Lnet/sourceforge/jpcap/capture/CaptureDevice;
 */
JNIEXPORT jobjectArray JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_lookupCaptureDevices0
  (JNIEnv *env, jclass packetCaptureClass)
{
	try
	{
		PcapInterfaces	interfaces;
		return buildCaptureDevices(env, interfaces.getDevices(), 0);
	}
	catch (const PcapDeviceLookupException &exc)
	{
	    throwException(env, DEVICE_LOOKUP_EXCEPTION, exc.getMessage());
		return 0;
	}
	catch (const PendingJavaException &)
	{
		return 0;
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    getLibVersion0
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_getLibVersion0
  (JNIEnv *env, jclass cls)
{
	return env->NewStringUTF(pcap_lib_version());
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    setLoopMode0
 * Signature: (JI)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_setLoopMode0
  (JNIEnv *env, jobject self, jlong handle, jint loopMode)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	pc->setLoopMode(loopMode == net_sourceforge_jpcap_capture_PacketCapture_LOOP_MODE);
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    writePacket0
 * Signature: (I[B)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_writePacket0
  (JNIEnv *env, jobject self, jlong handle, jobject bb)
{
	try
	{
		PacketCapture	*pc = (PacketCapture*)handle;
		JNUByteBuffer	jbb(bb);
		const void	*buf = jbb.getData();
		jlong	length = jbb.getLimit();
		struct pcap_pkthdr header;
		header.ts.tv_sec = header.ts.tv_usec = 0;
		header.len = header.caplen = (unsigned)length;
		pcap_dump((u_char*)pc->getDumper(), &header, (u_char*)buf);
	}
	catch (const PendingJavaException &)
	{
		;
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    captureToFile0
 * Signature: (II)I
 */
JNIEXPORT jint JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_captureToFile0
  (JNIEnv *env, jobject self, jlong handle, jint count)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	// Here we are using pcap_dump() as the pcap_loop() callback,
	// as described in 'man libpcap'.
	// This technique is about as efficient as you can get.
	int	captureCount = 0;
	int	rc;
	while (count < 0 || count > 0)
	{
		rc = pc->capture(count, pcap_dump, (u_char *)pc->getDumper());
		// printf("captureToFile0: rc=%d\n", rc); fflush(stdout);
		switch (rc)
		{
		case -1:
			// if the capture failed, throw an exception back to the java wrapper
			throwException(env, CAPTURE_PACKET_EXCEPTION, pcap_geterr(pc->getPD()));
			return rc;
		case 0:	// EOF
			if (pcap_file(pc->getPD()) != 0)
				// EOF on capture from file
				return captureCount;
			// fall through
		default:
			if (count > rc)
				count -= rc;
			captureCount += rc;
			continue;
		}
	}
	if (pcap_dump_flush(pc->getDumper()) < 0)
		throwException(env, IO_EXCEPTION, pcap_geterr(pc->getPD()));
	return rc == -2 ? 0 : rc < 0 ? rc : captureCount;
}

// Helper class for captureToNetwork0()
class InjectCallback : public PcapCallback
{
public:
	InjectCallback(JNIEnv *env, pcap_t *pd) : _env(env), _pd(pd){}

protected:
	void	handlePacket(const struct pcap_pkthdr *h, const unsigned char *sp)
	{
		if (_env->ExceptionCheck())
			return;
#ifndef WIN32
	    int	rc = pcap_inject(_pd, sp, h->caplen);
#else	// WIN32
		int	rc = pcap_sendpacket(_pd, sp, h->caplen);
#endif // WIN32
	//	printf("injectCallback len=%d caplen=%d secs=%s usecs=%d rc=%d\n", h->len, h->caplen, h->ts.tv_sec, h->ts.tv_usec, rc);
		if (rc < 0)
			throwException(_env, IO_EXCEPTION, pcap_geterr(_pd));
	}
private:
	JNIEnv	*_env;
	pcap_t	*_pd;
};

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    captureToNetwork0
 * Signature: (II)I
 */
JNIEXPORT jint JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_captureToNetwork0
  (JNIEnv *env, jobject self, jlong handle, jint count, jlong outHandle)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	PacketCapture	*out= (PacketCapture*)outHandle;
	InjectCallback	cb(env, out->getPD());

	// NB the cast to PcapCallback* is necessary so we don't lie to the thunk
	// about the type & address of the class object.
	int	rc = pc->capture(count, PcapCallback::pcap_handler_thunk, (u_char *)(PcapCallback*)&cb);
	if (rc < 0)
	{
		// if the capture failed, throw an exception back to the java wrapper
		throwException(env, CAPTURE_PACKET_EXCEPTION, pcap_geterr(pc->getPD()));
	}
	return rc == -2 ? 0 : rc;
}

/*
 * Class:     net_sourceforge_jpcap_capture_PacketCapture
 * Method:    setDirection0
 * Signature: (II)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_PacketCapture_setDirection0
  (JNIEnv *env, jobject self, jlong handle, jint direction)
{
#ifdef HAVE_SETDIRECTION
	PacketCapture	*pc = (PacketCapture*)handle;
	int	rc = pcap_setdirection(pc->getPD(), (pcap_direction_t)direction);
	if (rc < 0)
	{
		// if the capture failed, throw an exception back to the java wrapper
		throwException(env, IO_EXCEPTION, pcap_geterr(pc->getPD()));
	}
#else
	throwException(env, UNSUPPORTED_OPERATION_EXCEPTION, "setDirection");
#endif
}

// WinPcap.java support functions.

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap
 * Method:    isSupported0
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_isSupported0
  (JNIEnv *env, jclass cls)
{
#ifndef WIN32
	return JNI_FALSE;
#else
	return JNI_TRUE;
#endif
}

#ifdef WIN32
// Everything else is only present if it is supported, i.e. if it will compile!

// Helper class for captureToNetwork0()
// Notes: a pcap_send_queue can't be reused after calling pcap_sendqueue_transmit.
class SendQueueInjectCallback : public PcapCallback
{
public:
	SendQueueInjectCallback(JNIEnv *env, unsigned queueLength, bool sync, PacketCapture *out)
		throw(PcapSendQueueException)
		: _env(env),
		_queue(out, queueLength, sync)
	{
	}

	void	handlePacket(const pcap_pkthdr *h, const u_char *sp)
	{
		if (_env->ExceptionCheck())
			return;
		try
		{
			_queue.queue(h, sp);
		}
		catch (const PcapSendQueueException &exc)
		{
			throwException(_env, IO_EXCEPTION, exc.getMessage());
		}

	}

private:
	JNIEnv			*_env;
	PcapSendQueue	_queue;
};


/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap
 * Method:    captureToNetwork0
 * Signature: (III)I
 */
JNIEXPORT jint JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_captureToNetwork0
  (JNIEnv *env, jobject self, jlong handle, jint count, jlong outHandle)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	PacketCapture	*out = (PacketCapture*)outHandle;
	// NB this is allocated on the stack so its destructor is guaranteed to be called.
	try
	{
		SendQueueInjectCallback	h(env, DEFAULT_BUFFER_SIZE, 1, out);

		// NB the cast to PcapCallback* is necessary so we don't lie to the thunk
		// about the type & address of the class object.
		int	rc = pc->capture(count, PcapCallback::pcap_handler_thunk, (u_char *)(PcapCallback*)&h);
		if (rc < 0)
		{
			// if the capture failed, throw an exception back to the java wrapper
			throwException(env, CAPTURE_PACKET_EXCEPTION, pcap_geterr(pc->getPD()));
		}
		return rc == -2 ? 0 : rc;
	}
	catch (const PcapSendQueueException &exc)
	{
		throwException(env, IO_EXCEPTION, exc.getMessage());
		return 0;
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap
 * Method:    isFilteredOffline0
 * Signature: (Ljava/lang/String;[BI)Z
 */
JNIEXPORT jboolean JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_isFilteredOffline0
  (JNIEnv *env, jobject self, jlong filter, jobject bb)
{
	try
	{
		JNUByteBuffer	jbb(bb);
		const void	*data = jbb.getData();
		jlong	length = jbb.getLimit();
		struct bpf_program	*bpf = (struct bpf_program *)filter;
		pcap_pkthdr header;
		header.caplen = header.len = (unsigned)length;
		header.ts.tv_sec = header.ts.tv_usec = 0;
		int rc = pcap_offline_filter(bpf, &header, (u_char*)data);
		return rc;
	}
	catch (const PendingJavaException &)
	{
		return 0;
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap
 * Method:    lookupCaptureDevices0
 * Signature: ()[Lnet/sourceforge/jpcap/capture/CaptureDevice;
 */
JNIEXPORT jobjectArray JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_lookupCaptureDevices0
  (JNIEnv *env, jclass cls, jstring juri, int type)
{
	try
	{
		// Push a new local frame
		JNULocalFrame	frame;
		char ebuf[PCAP_ERRBUF_SIZE] = "";
		int	rc;
		char	source[PCAP_BUF_SIZE] = PCAP_SRC_IF_STRING;
		// char	*uri = const_cast<char*>(juri ? env->GetStringUTFChars(juri, 0) : 0);
		JNUString	uri(juri);

		TRACE((printf("JWinPcap.lookupCaptureDevices(): entered uri=%s type=%d\n", uri.getLength() ? uri.getUTFChars() : "", type)));
		if (uri.getLength())
		{
			char	host[PCAP_BUF_SIZE];
			char	port[PCAP_BUF_SIZE] = "";
			char	name[PCAP_BUF_SIZE];
			rc = pcap_parsesrcstr(uri, &type, host, port, name, ebuf);
			TRACE(printf(
				"JWinPcap.lookupCaptureDevices(): pcap_parsesrcstr rc=%d type=%d host=%s port=%s name=%s\n",
				rc,
				type,
				host,
				port,
				name
				));
			if (rc < 0)
			{
				throwException(env, DEVICE_LOOKUP_EXCEPTION, ebuf);
				return 0;
			}
			switch (type)
			{
			case PCAP_SRC_IFREMOTE:
			case PCAP_SRC_FILE:
			case PCAP_SRC_IFLOCAL:
				rc = pcap_createsrcstr(source, type, host, port, name, ebuf);
				break;
			}
			TRACE(printf("JWinPcap.lookupCaptureDevices() pcap_createsrcstr->source=%s rc=%d err=%s\n", source, rc, ebuf));
			if (rc < 0)
			{
				throwException(env, DEVICE_LOOKUP_EXCEPTION, ebuf);
				return 0;
			}
		}

		// etc as for pcap_findalldevs().
		PcapInterfaces	interfaces(source, NULL);
		jobjectArray devices = buildCaptureDevices(env, interfaces.getDevices(), 1);
		TRACE((printf("JWinPcap.lookupCaptureDevices(): exit\n")));
		return frame.pop(devices);
	}
	catch (const PcapDeviceLookupException &exc)
	{
		throwException(env, DEVICE_LOOKUP_EXCEPTION, exc.getMessage());
		return 0;
	}
	catch (const PendingJavaException &)
	{
		return 0;
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap
 * Method:    open0
 * Signature: (ILjava/lang/String;IIILjava/net/Authenticator;)V
 */
JNIEXPORT jint JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_open0
  (JNIEnv *env, jobject self, jlong handle, jstring jname, jint snaplen, jint flags, jint timeout, jstring jusername, jstring jpassword)
{
	TRACE((printf("JWinPcap.open(): entering\n")));
	try
	{
		// Declare a local frame
		JNULocalFrame	frame;
		PacketCapture	*pc = (PacketCapture*)handle;
		char ebuf[PCAP_ERRBUF_SIZE] = "";
		pcap_rmtauth	auth = {RPCAP_RMTAUTH_NULL, 0, 0};

		// Get the name
		char	source[PCAP_BUF_SIZE];
		// TODO vary this?
		JNUString	name(jname);
		sprintf(source, "%s", name.getUTFChars());
		TRACE((printf("source=%s\n", source)));

		JNUString	username(jusername);
		// Possible UTF problem here, original was Java char[]
		JNUString	password(jpassword);


		// Get the authentication
		if (jusername != 0)
		{
			auth.type = RPCAP_RMTAUTH_PWD;
			auth.username = const_cast<char*>(username.getUTFChars());
			auth.password = const_cast<char*>(password.getUTFChars());
		}

		pc->setPD(pcap_open(source, snaplen, flags, timeout, &auth, ebuf));
		if (pc->getPD() == 0)
		{
			throwException(env, IO_EXCEPTION, ebuf);
			return 0;
		}
		if (strlen(ebuf) > 0)
		{
			TRACE(printf("JWinPcap.open(): warning %s\n", ebuf));
		}
		TRACE((printf("JWinPcap.open(): exit\n")));
		return pcap_datalink(pc->getPD());
	}
	catch (const PendingJavaException &)
	{
		return 0;
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap
 * Method:    setBufferSize0
 * Signature: (II)I
 */
JNIEXPORT jint JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_setBufferSize0
  (JNIEnv *env, jobject self, jlong handle, jint size)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	int	rc = pcap_setbuff(pc->getPD(), size);
	if (rc < 0)
	    throwException(env, IO_EXCEPTION, pcap_geterr(pc->getPD()));
	return rc;
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap
 * Method:    setCaptureMode0
 * Signature: (II)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_setCaptureMode0
  (JNIEnv *env, jobject self, jlong handle, jint mode)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	if (pcap_setmode(pc->getPD(), mode) < 0)
	    throwException(env, ILLEGAL_ARGUMENT_EXCEPTION, pcap_geterr(pc->getPD()));
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap
 * Method:    setMinimumKernelCopy0
 * Signature: (II)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_setMinimumKernelCopy0
  (JNIEnv *env, jobject self, jlong handle, jint minimum)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	if (pcap_setmintocopy(pc->getPD(), minimum) < 0)
	    throwException(env, IO_EXCEPTION, pcap_geterr(pc->getPD()));
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap
 * Method:    getSamplingMethod0
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_getSamplingMethod0
  (JNIEnv *, jobject self, jlong handle)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	pcap_samp	*pcs = pcap_setsampling(pc->getPD());
	return pcs->method;
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap
 * Method:    getSamplingValue0
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_getSamplingValue0
  (JNIEnv *, jobject self, jlong handle)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	pcap_samp	*pcs = pcap_setsampling(pc->getPD());
	return pcs->value;
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap
 * Method:    setSamplingMethod0
 * Signature: (III)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_setSamplingMethod0
  (JNIEnv *env, jobject self, jlong handle, jint method, jint value)
{
	PacketCapture	*pc = (PacketCapture*)handle;
	pcap_samp	*pcs = pcap_setsampling(pc->getPD());
	pcs->method = method;
	pcs->value = value;
}

// JWinPcap.SendQueue support functions.

/*
 * Class:     Java_net_sourceforge_jpcap_capture_JWinPcap
 * Method:  sendQueueCreate0
 * Signature: (I)J
 */
JNIEXPORT jlong JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_sendQueueCreate0
  (JNIEnv *env, jclass cls, jlong handle, jint dim, jboolean sync)
{
//	TRACE((printf("JWinPcap.sendQueueCreate0(): entered dim=%d\n", dim)));
	PacketCapture	*pc = (PacketCapture*)handle;
	try
	{
		PcapSendQueue	*queue = new PcapSendQueue(pc, dim, sync);
//	TRACE((printf("JWinPcap.sendQueueCreate0(): exit queue=0x%8.8x\n", queue)));
		return (jlong)queue;
	}
	catch (const PcapSendQueueException &exc)
	{
		throwException(env, IO_EXCEPTION, exc.getMessage());
		return 0;
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap_SendQueue
 * Method:    sendQueueDestroy0
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_sendQueueDestroy0
  (JNIEnv *env, jobject self, jlong queue)
{
//	TRACE((printf("JWinPcap.sendQueueDestroy0(): entered\n")));
	delete (PcapSendQueue*)queue;
//	TRACE((printf("JWinPcap.sendQueueDestroy0(): entered\n")));
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap_SendQueue
 * Method:    sendQueueEnqueue0
 * Signature: (J[BI[BIJJ)V
 */
JNIEXPORT jint JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_sendQueueEnqueue0
  (JNIEnv *env, jobject self, jlong queue, jobject bb, jint sec, jint usec)
{
//	TRACE((printf("JWinPcap.sendQueueEnqueue0(): entered queue=0x%8.8x%8.8x\n", (int)(ptr >> 32), (int)ptr)));
//	TRACE((printf("JWinPcap.sendQueueEnqueue0(): entered\n")));
	try
	{
		JNUByteBuffer	jbb(bb);
		jlong	length = jbb.getLimit();
		// Get a pointer to the data to send.
		const void	*data = jbb.getData();
		// fprintf(stderr, "JWinPcap.send0: data=%8.8x length=%d\n", data, length);
		pcap_pkthdr	hdr;
		hdr.caplen = hdr.len = (unsigned)length;
		hdr.ts.tv_sec = sec;
		hdr.ts.tv_usec = usec;
		PcapSendQueue	*q = (PcapSendQueue*)queue;
		q->queue(&hdr, (u_char*)data);
		TRACE((printf("JWinPcap.sendQueueEnqueue0(): exit rc=OK\n")));
		return 0;
	}
	catch (const PcapSendQueueException &exc)
	{
		TRACE((printf("JWinPcap.sendQueueEnqueue0(): exception %s\n", exc.getMessage())));
		throwException(env, IO_EXCEPTION, exc.getMessage());
		return 0;
	}
	catch (const PendingJavaException &)
	{
		return 0;
	}
}

/*
 * Class:     net_sourceforge_jpcap_capture_JWinPcap_SendQueue
 * Method:    sendQueueTransmit0
 * Signature: (JZ)V
 */
JNIEXPORT jint JNICALL Java_net_sourceforge_jpcap_capture_JWinPcap_sendQueueTransmit0
  (JNIEnv *env, jobject self, jlong handle, jlong queue)
{
//	TRACE((printf("JWinPcap.sendQueueTransmit0(): entered instance=%d pc->getPD()=0x%8.8x queue=0x%8.8x\n", instance, pc->getPD(), ptr)));
	PacketCapture	*pc = (PacketCapture*)handle;
	PcapSendQueue	*q = (PcapSendQueue*)queue;
	try
	{
		q->flush();
		return 0;
	}
	catch (const PcapSendQueueException &exc)
	{
		throwException(env, IO_EXCEPTION, exc.getMessage());
		return 0;
	}
}

#endif

static const char rcsid[] =
  "$Id: jpcap.cpp,v 1.4 2008/10/06 15:08:18 rickwall Exp $";
