// $Id: PacketHandler.java,v 1.2 2008/08/18 05:34:36 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.capture;

import java.nio.ByteBuffer;


/**
 * This is the jpcap <b>internal</b> capture interface.
 * libpcap calls back the jpcap packet handler via this interface when packets arrive.
 * <p> 
 * If you are writing an application which captures packets from jpcap, 
 * you probably want to look at PacketCapture and PacketListener.
 * <p>
 * The scope of this class is package-private because it is only used internally
 * as a contract between jpcap.dll and PacketCaptureBase.
 *
 * @author Patrick Charles and Jonas Lehmann
 * @author Esmond Pitt
 * @version $Revision: 1.2 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/18 05:34:36 $
 */
abstract class PacketHandler
{
  /**
   * Handle an arriving packet.
   * 
   * @param length the length of the packet off of the wire.
   * @param caplen the number of bytes actually captured.
   * @param seconds the seconds component of the timestamp.
   * @param useconds the microseconds component of the timestamp.
   * @param data the contents of the captured packet.
   */
  abstract void handlePacket(int length, int caplen, int seconds, int useconds, 
                    ByteBuffer bb);
}
