package net.sourceforge.jpcap.util;

import java.io.EOFException;

import net.sourceforge.jpcap.capture.PacketCapture;
import net.sourceforge.jpcap.net.IPPacket;
import net.sourceforge.jpcap.net.Packet;
import net.sourceforge.jpcap.net.TCPPacket;

public class AnalysisPcap {

	private static int nonTCPPackets = 0;
	private static int totalPackets = 0;
	public static long totalNumBytes = 0;
	public static long currentOffset = 0;
	public static int numPackets[] = new int[8];
	
	static PacketCapture pcapOffline = null;
  
	public static void loadPcap(String pcapFile) throws Exception {
		pcapOffline = new PacketCapture();
		pcapOffline.setSnaplen(65000);
		pcapOffline.openOffline(pcapFile);
	}
	
	/**
	 * This class will scan through a PCAP and identify packets that do not belong to a valid stream.
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		// We will want to identify the following 
		// Streams with TCP SYN and PROPER FIN-ACK or RST. - TABLE
		// Streams without TCP SYN - TABLE
		// Streams without FIN-ACK or RST. - Table
		// Non-TYPE packets. - just a counter.

		// First we need to load the PCAP.
		// Second we need to find all TCP SYN connections.
		// Next we need to walk through the pcap file counting up packets for each SYN created connection.
		// We will need to note TCP packets that do not have a valid SYN matching.
		// We then need to note non-TCP Packets.
		// Finally we need to note TCP connections that were properly closed to ones that aren't.
		
		loadPcap(args[0]);
		
		Packet offlinePacket = null;
		PacketStreams h = new PacketStreams();
		IPPacket t = null;
		int x = 0, between = 100;
	
		while (true) {
			try {
				offlinePacket = (Packet) pcapOffline.captureNext();
			} catch (EOFException g) {
				offlinePacket=null;
			}
			if (offlinePacket == null)
				break;
			
			totalPackets++;
			//System.out.println("Total Packets : " + totalPackets);
			if(totalNumBytes < between ) {
				numPackets[x]++;
			} else {
				between = between*10;
				if (x < numPackets.length)
					x++;
			}
	
			totalNumBytes = totalNumBytes + offlinePacket.getPacketLength();
			
			t = (IPPacket) offlinePacket;
			if (t.getEthernetProtocol() == 2048 && t.getIPProtocol() == 6) { // IP Protocol
				h.add((TCPPacket)t);
			} else {
				nonTCPPackets++;
			}
		}
		
		int numValidStreams = 0, numInvalidStreams = 0;
		StringBuffer invalidStreams = new StringBuffer(), validStreams = new StringBuffer();
		
		for (TCPSession[] sessions : h.values() )
			for (TCPSession session : sessions) {
				if (session.isSYN() && (session.isClientClosed() || session.isServerClosed())) {
					validStreams.append("\n"
						+ Convert.ipToString(session.getDstIP()) + ":"
						+ session.getDstPort() + " "
						+ Convert.ipToString(session.getSrcIP()) + ":"
						+ session.getSrcPort());
					numValidStreams++;			
				} else {
					invalidStreams.append("\n"
						+ Convert.ipToString(session.getDstIP()) + ":"
						+ session.getDstPort() + " "
						+ Convert.ipToString(session.getSrcIP()) + ":"
						+ session.getSrcPort());
					numInvalidStreams++;
				}
			}
		
		if(numInvalidStreams>100)
			System.out.println("Invalid Streams are : too many to print, more than 100");
		else if (numInvalidStreams > 0)
			System.out.println("Invalid Streams are : " + invalidStreams.toString());
		if(numValidStreams>100)
			System.out.println("\n\nValid Streams are : too many to print, more than 100");
		else if (numValidStreams > 0)
			System.out.println("\n\nValid Streams are : " + validStreams.toString());
		
		System.out.print("\n");
		
		System.out.println("Number of Valid Streams (include SYN and RST or FIN ACK) : " + numValidStreams);
		System.out.println("Number of invalid Streams (missing either SYN and RST or FIN ACK) : " + numInvalidStreams);
		
		/*
		 * Displaying the results
		 */
		System.out.println("Total Packets : " + totalPackets);
		System.out.println("Total non-TCP Packets : " + nonTCPPackets);
		System.out.println("Number of TCP Streams : " + (numInvalidStreams + numValidStreams));
		System.out.println("Total number of bytes : " + totalNumBytes);
		
		System.out.print("\n");
		between = 10;
		System.out.println("To send the whole file at 200meg/s it would take " +  ((float)totalNumBytes*8/(200*1024*1024)) + " seconds");
		for (int y : numPackets) {
			System.out.println("Packets between " + (between==10 ? 0:between) + " - " + between*10 + " : (" + (between==10 ? 0:between*8/(200*1024*1024)) + "-" + (float)between*10*8/(200*1024*1024) + ") " + y);
			between = between*10;
		}
	}
}
