// $Id: Timeval.java,v 1.2 2008/08/18 05:34:48 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.util;

import java.util.Date;
import java.io.Serializable;
import java.text.DecimalFormat;
import java.text.Format;


/**
 * POSIX.4 timeval for Java.
 * <p>
 * Container for java equivalent of c's struct timeval.
 *
 * @author Patrick Charles and Jonas Lehmann
 * @version $Revision: 1.2 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/18 05:34:48 $
 */
public class Timeval implements Serializable
{
	static final long serialVersionUID = -523538551667488015L;
	
	// Fix for 1590171
	private static Format	format = new DecimalFormat("000000");
	
  public Timeval(long seconds, int microseconds) {
    this.seconds = seconds;
    this.microseconds = microseconds;
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append(seconds);
    sb.append('.');
	// Fix for 1590171
    sb.append(format.format(microseconds));
    sb.append('s');
    return sb.toString();
  }

  /**
   * Convert this timeval to a java Date.
   */
  public Date getDate() {
    return new Date(seconds * 1000 + microseconds/1000);
  }

  public long getSeconds() {
    return seconds;
  }

  public int getMicroSeconds() {
    return microseconds;
  }

  long seconds;
  int microseconds;

  private static final String _rcsid = 
  "$Id: Timeval.java,v 1.2 2008/08/18 05:34:48 esmondpitt Exp $";
}
